import datetime as dt
import gzip
import json
from enum import Enum
from functools import partial

import httpx
import interloper as il
import pandas as pd
from interloper_pandas import DataFrameNormalizer
from pydantic_settings import SettingsConfigDict
from tenacity import (
    RetryCallState,
    Retrying,
    retry,
    retry_if_exception,
    retry_if_result,
    stop_after_attempt,
    stop_after_delay,
    wait_exponential,
    wait_incrementing,
)

from interloper_assets.amazon_ads import constants, schemas


class AmazonAdsAPILocation(Enum):
    NORTH_AMERIC = "NA"
    EUROPE = "EU"
    FAR_EAST = "FE"

    @property
    def api_url(self) -> str:
        return {
            AmazonAdsAPILocation.EUROPE: "https://advertising-api-eu.amazon.com",
            AmazonAdsAPILocation.FAR_EAST: "https://advertising-api-fe.amazon.com",
            AmazonAdsAPILocation.NORTH_AMERIC: "https://advertising-api.amazon.com",
        }[self]

    @property
    def auth_url(self) -> str:
        return {
            AmazonAdsAPILocation.EUROPE: "https://api.amazon.co.uk",
            AmazonAdsAPILocation.FAR_EAST: "https://api.amazon.co.jp",
            AmazonAdsAPILocation.NORTH_AMERIC: "https://api.amazon.com",
        }[self]


class AmazonAdsConfig(il.Config):
    location: str
    profile_id: str
    client_id: str
    client_secret: str
    refresh_token: str

    model_config = SettingsConfigDict(env_prefix="amazon_ads_")


@il.source(
    config=AmazonAdsConfig,
    tags=["Advertising"],
    normalizer=DataFrameNormalizer(flatten_max_level=1),
)
class AmazonAds:
    config: AmazonAdsConfig
    location: AmazonAdsAPILocation
    client: il.RESTClient

    def __init__(self, config: AmazonAdsConfig) -> None:
        self.location = AmazonAdsAPILocation(config.location)
        self.client = il.RESTClient(
            self.location.api_url,
            auth=il.OAuth2RefreshTokenAuth(
                base_url=self.location.auth_url,
                token_endpoint="/auth/o2/token",
                client_id=config.client_id,
                client_secret=config.client_secret,
                refresh_token=config.refresh_token,
            ),
        )
        self.client.base_url = self.location.api_url
        self.client.headers.update({"Amazon-Advertising-API-ClientId": config.client_id})

    # ------------------------------------------------------------------
    # HELPERS
    # ------------------------------------------------------------------

    def request_report(
        self,
        report_type_id: str,
        profile_id: str,
        ad_product: str,
        group_by: list[str],
        columns: list[str],
        start_date: dt.date,
        end_date: dt.date,
    ) -> dict:
        response = self.client.post(
            "/reporting/reports",
            headers={
                "Amazon-Advertising-API-Scope": profile_id,
                "Content-Type": "application/vnd.createasyncreportrequest.v3+json",
            },
            json={
                "name": f"{report_type_id} | {start_date.isoformat()} - {end_date.isoformat()}",
                "startDate": start_date.isoformat(),
                "endDate": end_date.isoformat(),
                "configuration": {
                    "reportTypeId": report_type_id,
                    "adProduct": ad_product,
                    "groupBy": group_by,
                    "columns": columns,
                    "timeUnit": "DAILY",
                    "format": "GZIP_JSON",
                },
            },
        )
        response.raise_for_status()
        return response.json()

    def get_report_status(self, profile_id: str, report_id: str) -> dict:
        response = self.client.get(
            f"/reporting/reports/{report_id}",
            headers={
                "Amazon-Advertising-API-Scope": profile_id,
                "Content-Type": "application/vnd.createasyncreportrequest.v3+json",
            },
        )
        response.raise_for_status()
        return response.json()

    def download_report(self, url: str) -> list[dict]:
        # Use httpx.get directly to not use the client's auth
        response = httpx.get(url, timeout=None)
        response.raise_for_status()

        gzip_content = response.content
        json_content = gzip.decompress(gzip_content)

        return json.loads(json_content)

    def _log_retry(self, context: il.ExecutionContext, retry_state: RetryCallState) -> None:
        if retry_state.outcome is None:
            raise RuntimeError("log called before outcome was set")

        if retry_state.next_action is None:
            raise RuntimeError("log called before next_action was set")

        if retry_state.outcome.failed:
            ex = retry_state.outcome.exception()
            context.logger.debug(f"Retrying in {retry_state.next_action.sleep}. Raised {ex.__class__.__name__}: {ex}.")
            if isinstance(ex, httpx.HTTPStatusError) and ex.response.status_code == 429:
                context.logger.debug(f"Retry-After response header: {ex.response.headers.get('Retry-After')}")
        else:
            context.logger.debug(
                f"Retrying in {retry_state.next_action.sleep}. Returned {retry_state.outcome.result()}"
            )

    @retry(stop=stop_after_attempt(5))
    def _clear_auth(self) -> None:
        auth = self.client.auth
        if isinstance(auth, il.OAuth2RefreshTokenAuth):
            # Clear the token to force a new token to be acquired
            auth.clear_token()

    def request_and_download_report(
        self,
        context: il.ExecutionContext,
        report_type_id: str,
        profile_id: str,
        ad_product: str,
        group_by: list[str],
        columns: list[str],
        start_date: dt.date,
        end_date: dt.date,
        max_request_delay: int | dt.timedelta = dt.timedelta(hours=1),
        max_wait_delay: int | dt.timedelta = dt.timedelta(hours=2),
        max_download_delay: int | dt.timedelta = dt.timedelta(minutes=10),
    ) -> list[dict]:
        ConfiguredRetrying = partial(
            Retrying,
            wait=wait_exponential(multiplier=10, max=60),
            before=lambda _: self._clear_auth(),
            before_sleep=partial(self._log_retry, context),
            reraise=True,
        )

        for attempt in ConfiguredRetrying(
            stop=stop_after_delay(max_request_delay),
            wait=wait_incrementing(start=60, increment=dt.timedelta(minutes=5), max=dt.timedelta(minutes=15)),
        ):
            with attempt:
                context.logger.info(
                    f"Requesting {report_type_id} report for profile {profile_id} "
                    f"(attempt {attempt.retry_state.attempt_number})"
                )
                data = self.request_report(
                    report_type_id, profile_id, ad_product, group_by, columns, start_date, end_date
                )

        report_id = data["reportId"]
        context.logger.info(f"Report id: {report_id} for profile {profile_id}")

        # Wait for report
        for attempt in ConfiguredRetrying(
            stop=stop_after_delay(max_wait_delay),
            retry=retry_if_result(lambda status: status != "COMPLETED")
            | retry_if_exception(lambda e: issubclass(e.__class__, httpx.HTTPError)),
        ):
            with attempt:
                context.logger.info(f"Waiting for report {report_id} (profile {profile_id}) ")
                response = self.get_report_status(profile_id, report_id)

            if attempt.retry_state.outcome and not attempt.retry_state.outcome.failed:
                attempt.retry_state.set_result(response["status"])

        report_url = response["url"]
        context.logger.info(f"Report URL: {report_url} for profile {profile_id}")

        for attempt in ConfiguredRetrying(
            stop=stop_after_delay(max_download_delay),
        ):
            with attempt:
                context.logger.info(f"Downloading report (attempt {attempt.retry_state.attempt_number})")
                report = self.download_report(report_url)

        return report

    # ------------------------------------------------------------------
    # ASSETS
    # ------------------------------------------------------------------

    @il.asset(
        schema=schemas.Profiles,
        tags=["Entity"],
    )
    def profiles(self) -> pd.DataFrame:
        response = self.client.get("/v2/profiles")
        response.raise_for_status()
        return pd.DataFrame(response.json())

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_advertised_products(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of advertised products including clicks, impressions, sales, purchases, and attributed sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spAdvertisedProduct",
            group_by=["advertiser"],
            columns=constants.PRODUCTS_ADVERTISED_PRODUCT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_campaigns(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Campaign performance for advertised products including clicks, impressions, sales, purchases, and attributed
        sales.
        """
        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spCampaigns",
            group_by=["campaign", "adGroup"],
            columns=constants.PRODUCTS_CAMPAIGN_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_search_terms(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of search terms in advertising campaigns including click-through rate, clicks, cost,
        impressions, purchases, and sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spSearchTerm",
            group_by=["searchTerm"],
            columns=constants.PRODUCTS_SEARCH_TERM_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_targeting(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of advertising campaigns based on targeting criteria including click-through rate, clicks, cost,
        impressions, purchases, and sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spTargeting",
            group_by=["targeting"],
            columns=constants.PRODUCTS_TARGETING_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_purchased_products(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of advertised products including sales, purchases, SKU, and units sold."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spPurchasedProduct",
            group_by=["asin"],
            columns=constants.PRODUCTS_PURCHASED_PRODUCT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def products_gross_and_invalid_traffic(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Impact of gross and invalid traffic on advertising campaigns including clicks, impressions, and traffic
        quality indicators.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_PRODUCTS",
            report_type_id="spGrossAndInvalids",
            group_by=["campaign"],
            columns=constants.PRODUCTS_GROSS_AND_INVALID_TRAFFIC_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_campaigns(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Display advertising campaign performance including clicks, impressions, purchases, sales, viewability, and
        attributed sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdCampaigns",
            group_by=["campaign"],
            columns=constants.DISPLAY_CAMPAIGN_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_advertised_products(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of advertised products within display ad campaigns including clicks, impressions, purchases, and
        sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdAdvertisedProduct",
            group_by=["advertiser"],
            columns=constants.DISPLAY_ADVERTISED_PRODUCT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_purchased_products(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of promoted products within display ad campaigns including conversions, sales, purchases, and
        attributed sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdPurchasedProduct",
            group_by=["asin"],
            columns=constants.DISPLAY_PURCHASED_PRODUCT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_targeting(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad campaign performance based on targeting criteria including clicks, impressions, purchases, and sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdTargeting",
            group_by=["targeting"],
            columns=constants.DISPLAY_TARGETING_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_gross_and_invalid_traffic(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Traffic quality for display campaigns including total clicks, impressions, and traffic quality indicators."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdGrossAndInvalids",
            group_by=["campaign"],
            columns=constants.DISPLAY_GROSS_AND_INVALID_TRAFFIC_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def display_ad_groups(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of ad groups within display campaigns including clicks, impressions, purchases, sales, and
        attributed sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_DISPLAY",
            report_type_id="sdAdGroup",
            group_by=["adGroup"],
            columns=constants.DISPLAY_AD_GROUP_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_campaigns(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Brand promotion campaign performance including impressions, clicks, conversions, and attributed sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbCampaigns",
            group_by=["campaign"],
            columns=constants.BRANDS_CAMPAIGN_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_ads(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Individual ad performance within brand promotion campaigns including impressions, clicks, conversions, and
        attributed sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbAds",
            group_by=["ads"],
            columns=constants.BRANDS_AD_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_search_terms(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Search term performance in brand campaigns including clicks, impressions, purchases, and sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbSearchTerm",
            group_by=["searchTerm"],
            columns=constants.BRANDS_SEARCH_TERM_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_targeting(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad performance based on targeting criteria in brand campaigns including clicks, impressions, purchases, and
        sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbTargeting",
            group_by=["targeting"],
            columns=constants.BRANDS_TARGETING_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_purchased_products(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Products purchased through brand campaigns including number of purchases, sales, and attributed sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbPurchasedProduct",
            group_by=["purchasedAsin"],
            columns=constants.BRANDS_PURCHASED_PRODUCT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_gross_and_invalid_traffic(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Traffic quality for brand campaigns including total clicks, impressions, and traffic quality indicators."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbGrossAndInvalids",
            group_by=["campaign"],
            columns=constants.BRANDS_GROSS_AND_INVALID_TRAFFIC_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_placements(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Campaign performance by ad placement including clicks, impressions, purchases, and sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbCampaignPlacement",
            group_by=["campaignPlacement"],
            columns=constants.BRANDS_PLACEMENT_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def brands_ad_groups(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad group performance within brand campaigns including impressions, clicks, conversions, and attributed
        sales.
        """

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_BRANDS",
            report_type_id="sbAdGroup",
            group_by=["adGroup"],
            columns=constants.BRANDS_AD_GROUP_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def television_campaigns(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Television campaign performance including clicks, impressions, purchases, and sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_TELEVISION",
            report_type_id="stCampaigns",
            group_by=["campaign"],
            columns=constants.TELEVISION_CAMPAIGN_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)

    @il.asset(
        partitioning=il.TimePartitionConfig(column="date"),
        tags=["Report"],
    )
    def television_targeting(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Television campaign performance by targeting including clicks, impressions, purchases, and sales."""

        data = self.request_and_download_report(
            context,
            profile_id=self.config.profile_id,
            ad_product="SPONSORED_TELEVISION",
            report_type_id="stTargeting",
            group_by=["targeting"],
            columns=constants.TELEVISION_TARGETING_METRICS,
            start_date=context.partition_date,
            end_date=context.partition_date,
        )
        return pd.DataFrame(data)



import interloper as il
import pandas as pd


@il.source(tags=["Social"])
class LinkedinOrganic:
    @il.asset(tags=["Report"])
    def page(self, context: il.ExecutionContext) -> pd.DataFrame:
        """LinkedIn organic page statistics."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def posts(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance metrics for LinkedIn posts."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def post(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance metrics for individual LinkedIn posts."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def followers(self, context: il.ExecutionContext) -> pd.DataFrame:
        """LinkedIn follower counts by various dimensions."""

        raise NotImplementedError



import datetime as dt

from interloper.schema import AssetSchema
from pydantic import Field


class Campaigns(AssetSchema):
    """
    The Campaigns report provides insights into the performance and financial outcomes of advertising campaigns.
    It includes key metrics such as click rates, conversion rates, revenue from various client activities such as
    clicks, impressions, leads, and sales, as well as campaign and media information.
    """

    agent_id: int = Field(description="The unique identifier for the agent")
    camp_id: int = Field(description="The unique identifier for the campaign")
    camp_title: str = Field(description="The title of the campaign")
    campaign_manager: str = Field(description="The name of the campaign manager")
    campaign_manager_id: int = Field(description="The unique identifier for the campaign manager")
    click_nun: int = Field(description="The number of clicks from unique visitors")
    click_un: int = Field(description="The number of clicks from unique sources")
    client_click_rev: float = Field(description="Revenue generated from client clicks")
    client_impr_rev: float = Field(description="Revenue generated from client impressions")
    client_lead_price: float = Field(description="Price of leads for the client")
    client_lead_rev: float = Field(description="Revenue generated from client leads")
    client_sale_rev: float = Field(description="Revenue generated from client sales")
    client_subtotal: float = Field(description="Total revenue generated from client activities")
    company_name: str = Field(description="The name of the company")
    cpc: float = Field(description="Cost per click")
    cr: float = Field(description="Conversion rate")
    currency: str = Field(description="The currency used for transactions")
    currency_code: str = Field(description="The currency code used")
    currency_id: int = Field(description="The unique identifier for the currency")
    currency_rate: float = Field(description="The exchange rate of the currency")
    date: dt.date = Field(description="The date of the record")
    date_from: dt.date = Field(description="The starting date for a time period")
    date_max: dt.date = Field(description="The maximum date in a time period")
    date_min: dt.date = Field(description="The minimum date in a time period")
    date_to: dt.date = Field(description="The ending date for a time period")
    default_banner: str = Field(description="The default banner for the campaign")
    from_date: dt.date = Field(description="The starting date of an event or activity")
    leads: int = Field(description="The number of leads generated")
    media_id: int = Field(description="The unique identifier for the media")
    medianame: str = Field(description="The name of the media")
    mediatype: str = Field(description="The type of media")
    monthyear: str = Field(description="The combination of month and year")
    pcr: float = Field(description="Partial conversion rate")
    pending_rev: float = Field(description="Revenue that is pending or not yet realized")
    primary_category: int = Field(description="The primary category identifier")
    primary_category_name: str = Field(description="The name of the primary category")
    publisher_manager: str = Field(description="The name of the publisher manager")
    publisher_manager_id: int = Field(description="The unique identifier for the publisher manager")
    sales: float = Field(description="The number of sales")
    sales_amount: float = Field(description="The total amount of sales")
    stamp: dt.date = Field(description="The timestamp of the record")
    to_date: dt.date = Field(description="The ending date of an event or activity")
    week_to: dt.date = Field(description="The ending week of a time period")
    weekyear: str = Field(description="The combination of week and year")
    year: dt.date = Field(description="The year of the record")
    year_from: dt.date = Field(description="The starting year of a time period")
    year_to: dt.date = Field(description="The ending year of a time period")



import interloper as il
import pandas as pd


@il.source(tags=["Advertising"])
class AmazonSellingPartner:
    @il.asset(tags=["Report"])
    def vendor_sales_retail_manufacturing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Retail performance of manufactured products including ordered units, revenue, cost of goods sold, and profit
        margins.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_sales_retail_sourcing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Retail performance of sourced products including shipped units, revenue, cost of goods sold, and profit
        margins.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_sales_business_manufacturing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Sales performance and manufacturing-related metrics
        including revenue, units sold, and manufacturing costs.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_sales_business_sourcing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Performance of sourced products including shipped units, revenue, cost of goods sold, and profit margins."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_inventory_retail_manufacturing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Inventory management and performance including inventory levels, stock movements, and inventory costs."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_inventory_retail_sourcing(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Sourcing aspect of inventory management including inventory cost, units, lead time, and stock levels."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_forecasting_retail(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Forecasted demand for products including mean forecast units, 70th percentile forecast, and forecast
        accuracy.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_traffic(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Product visibility and engagement on the Amazon marketplace including glance views."""

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def vendor_net_pure_product_margin(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Net pure product margin insights for products."""

        raise NotImplementedError



import datetime

from interloper.schema import AssetSchema
from pydantic import Field


class BrandsAdGroups(AssetSchema):
    """
    The Brands Ad Group report provides insights into the performance of ad groups within Amazon Ads campaigns for
    brand promotion. It includes key metrics such as impressions, clicks, conversions, add-to-cart actions, sales,
    purchases, viewability rate, and various new-to-brand metrics.
    """

    ad_group_id: int | None = Field(..., description="The ID of the ad group")
    ad_group_name: str | None = Field(..., description="The name of the ad group")
    ad_status: str | None = Field(..., description="The status of the ad")
    add_to_cart: float | None = Field(..., description="The number of items added to cart")
    add_to_cart_clicks: float | None = Field(..., description="The number of clicks resulting in adding to cart")
    add_to_cart_rate: float | None = Field(..., description="The rate of adding to cart")
    branded_searches: float | None = Field(..., description="The number of branded searches")
    branded_searches_clicks: float | None = Field(
        ..., description="The number of clicks resulting from branded searches"
    )
    campaign_budget_amount: float | None = Field(..., description="The budget amount for the campaign")
    campaign_budget_currency_code: str | None = Field(..., description="The currency code for the campaign budget")
    campaign_budget_type: str | None = Field(..., description="The type of budget for the campaign")
    campaign_id: int | None = Field(..., description="The ID of the campaign")
    campaign_name: str | None = Field(..., description="The name of the campaign")
    campaign_status: str | None = Field(..., description="The status of the campaign")
    clicks: float | None = Field(..., description="The number of clicks")
    cost: float | None = Field(..., description="The cost of the ad")
    cost_type: str | None = Field(..., description="The type of cost")
    date: datetime.date | None = Field(..., description="The date of the data entry")
    detail_page_views: float | None = Field(..., description="The number of views of the detail page")
    detail_page_views_clicks: float | None = Field(
        ..., description="The number of clicks resulting in detail page views"
    )
    ecp_add_to_cart: float | None = Field(..., description="The ECP (effective cost per) add to cart")
    impressions: float | None = Field(..., description="The number of impressions")
    new_to_brand_detail_page_view_rate: float | None = Field(
        ..., description="The rate of detail page views from new-to-brand customers"
    )
    new_to_brand_detail_page_views: float | None = Field(
        ..., description="The number of detail page views from new-to-brand customers"
    )
    new_to_brand_detail_page_views_clicks: float | None = Field(
        ..., description="The number of clicks resulting in detail page views from new-to-brand customers"
    )
    new_to_brand_ecp_detail_page_view: float | None = Field(
        ..., description="The ECP (effective cost per) detail page view from new-to-brand customers"
    )
    new_to_brand_purchases: float | None = Field(..., description="The number of new-to-brand purchases")
    new_to_brand_purchases_clicks: float | None = Field(
        ..., description="The number of clicks resulting in new-to-brand purchases"
    )
    new_to_brand_purchases_percentage: float | None = Field(
        ..., description="The percentage of purchases from new-to-brand customers"
    )
    new_to_brand_purchases_rate: float | None = Field(
        ..., description="The rate of purchases from new-to-brand customers"
    )
    new_to_brand_sales: float | None = Field(..., description="The total sales amount from new-to-brand customers")
    new_to_brand_sales_clicks: float | None = Field(
        ..., description="The number of sales clicks from new-to-brand customers"
    )
    new_to_brand_sales_percentage: float | None = Field(
        ..., description="The percentage of sales from new-to-brand customers"
    )
    new_to_brand_units_sold: float | None = Field(
        ..., description="The number of units sold from new-to-brand customers"
    )
    new_to_brand_units_sold_clicks: float | None = Field(
        ..., description="The number of clicks resulting in units sold from new-to-brand customers"
    )
    new_to_brand_units_sold_percentage: float | None = Field(
        ..., description="The percentage of units sold from new-to-brand customers"
    )
    purchases: float | None = Field(..., description="The number of purchases")
    purchases_clicks: float | None = Field(..., description="The number of clicks resulting in purchases")
    purchases_promoted: float | None = Field(..., description="The number of promoted purchases")
    sales: float | None = Field(..., description="The total sales amount")
    sales_clicks: float | None = Field(..., description="The number of sales clicks")
    sales_promoted: float | None = Field(..., description="The total promoted sales amount")
    units_sold: float | None = Field(..., description="The number of units sold")
    units_sold_clicks: float | None = Field(..., description="The number of clicks resulting in units sold")
    video_5_second_view_rate: float | None = Field(..., description="The rate of video 5-second views")
    video_5_second_views: float | None = Field(..., description="The number of video 5-second views")
    video_complete_views: float | None = Field(..., description="The number of complete video views")
    video_first_quartile_views: float | None = Field(..., description="The number of video first quartile views")
    video_midpoint_views: float | None = Field(..., description="The number of video midpoint views")
    video_third_quartile_views: float | None = Field(..., description="The number of video third quartile views")
    video_unmutes: float | None = Field(..., description="The number of video unmutes")
    viewability_rate: float | None = Field(..., description="The rate of viewability for the ad")



from interloper.schema import AssetSchema
from pydantic import Field


class CampaignsByCity(AssetSchema):
    """
    The Campaigns by City report provides insights into campaign performance segmented by cities.
    It includes key metrics such as conversion rates, the number of conversions, and user demographics
    such as browser usage, device type, operating system, and geographical coordinates (latitude and longitude).
    """

    agent_id: int = Field(description="The ID of the agent")
    browser: str = Field(description="The browser used")
    camp_id: int = Field(description="The ID of the campaign")
    camp_title: str = Field(description="The title of the campaign")
    city: str = Field(description="The city")
    conversion_pct: int = Field(description="The conversion percentage")
    conversions: int = Field(description="The number of conversions")
    coordinates_lat: float = Field(description="The latitude coordinates")
    coordinates_lng: float = Field(description="The longitude coordinates")
    country: str = Field(description="The country")
    device_model: str = Field(description="The device model")
    device_type: str = Field(description="The device type")
    os: str = Field(description="The operating system")



import interloper as il
import pandas as pd


@il.source(tags=["Advertising"])
class TiktokAds:
    @il.asset(tags=["Report"])
    def ads(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad performance in TikTok campaigns including clicks, conversion value, conversion rate, impressions, and
        spend.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def ads_by_country(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad performance by country including clicks, conversion value, conversion rate, cost per conversion,
        impressions, and spend.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def ads_by_age_gender(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad performance by age and gender demographics including clicks, conversions, conversion rate, impressions,
        and spend.
        """

        raise NotImplementedError

    @il.asset(tags=["Report"])
    def ads_by_platform(self, context: il.ExecutionContext) -> pd.DataFrame:
        """Ad performance by platform including clicks, conversion value, conversion rate, cost per conversion,
        impressions, and spend.
        """

        raise NotImplementedError
